import os

from _event import SceneEventConsumer
from _renderer import Renderer
from _color_store import LrColorToutStore
from _models import TrScene

from edge.lane_detector import LaneDetector
from edge import eventlog
from edge.image.image import ImRgba, LiveImEngine, LiveImArgs, ImLoader

from edge.image.types import ImFooterVars

from edge.service import tz, gpsd, react, about


from ._args import KernelArgs, ImLoaderArgs
from ._stats import init_service as init_monitoring, fini_service as fini_monitoring, IStatService
from .tl import init_service, fini_service, ITlService


class _Services:
    def __init__(self, tz_service, gps_service, stat_service, react_service, about_service, tl_service):
        """
        :type tz_service: tz.ITzService
        :type gps_service: gpsd.IGpsdService
        :type stat_service: IStatService
        :type react_service: react.IReactService
        :type about_service: about.IAboutService
        :type tl_service : ITlService
        """
        self.tz = tz_service
        self.gps = gps_service
        self.stat = stat_service
        self.react = react_service
        self.about = about_service
        self.tl = tl_service


class BaseKernel:
    def __init__(self, event_consumer, image_renderer, im_engine, services, im_loader_args,
                 twisted_reactor):
        """
        :type event_consumer: SceneEventConsumer
        :type image_renderer: Renderer
        :type services: _Services
        :type im_loader_args: ImLoaderArgs
        :type twisted_reactor: twisted.internet.interfaces.IReactorTCP
        """
        self._event_consumer = event_consumer
        self._image_renderer = image_renderer

        self._im_engine = im_engine
        self._services = services
        self._services.react.set_handler(self.handle_react)

        self._im_loader = ImLoader(im_loader_args.dh, im_loader_args.image_wh[0], im_loader_args.image_wh[1])
        twisted_reactor.addSystemEventTrigger('before', 'shutdown', self._im_loader.destroy)
        self._rgba_image = ImRgba()

        #
        self._react_ts = 0

    # noinspection PyUnusedLocal
    def handle_react(self, react_name, react_jnode):
        """
        :type react_name: str
        :type react_jnode: dict
        """
        # noinspection PyBroadException
        tr_scene = TrScene.unpack(react_jnode)
        try:
            self._im_loader.load_ref(tr_scene.im_ref, self._rgba_image)
        except ValueError as x:
            eventlog.log_e('fail: load image {} <{}> '.format(tr_scene.im_ref.url, x.message))
            return

        react_ts = tr_scene.im_ref.ts
        if react_ts > self._react_ts:
            self._react_ts = react_ts

            origin = self._services.about.get_origin(self._rgba_image.ts)
            im_vars = ImFooterVars(
                self._services.tz, self._services.about, self._rgba_image.ts, defs={},
                geolocation=origin.geolocation, location=origin.location)
            self._image_renderer.render_tr_scene(tr_scene, self._rgba_image, self._services.tl)
            self._im_engine.mk_live(self._rgba_image, im_vars)
            self._event_consumer.consume_scene(react_ts, self._rgba_image)
            self._services.stat.wh.im_size(self._rgba_image.w, self._rgba_image.h, self._rgba_image.dh)


class Kernel(BaseKernel):
    def __init__(self, twisted_reactor, args):
        """
        :type args: KernelArgs
        """
        lane_detector = LaneDetector(
            args.lane_detector_args.id_to_lane_list,
            args.lane_detector_args.img_width,
            args.lane_detector_args.img_height)

        react_service = react.init_service(twisted_reactor, args.react.endpoints)

        tz_service = tz.init_service(args.tz)
        gpsd_service = gpsd.init_service(twisted_reactor, args.gpsd)
        stat_service = init_monitoring(twisted_reactor, args.monitoring)

        about_service = about.init_service(args.about, gpsd_service)

        tl_service = init_service(twisted_reactor, args.tl)

        services = _Services(tz_service, gpsd_service, stat_service, react_service, about_service, tl_service)

        event_consumer = SceneEventConsumer.create(twisted_reactor, args.scene_event, args.image_store)

        color_tout_store = LrColorToutStore(twisted_reactor, args.lr_store_tout)

        renderer = Renderer(args.renderer, color_tout_store, lane_detector)

        live_args = LiveImArgs(args.footer)
        im_engine = LiveImEngine(live_args)

        BaseKernel.__init__(self, event_consumer, renderer, im_engine, services, args.im_loader,
                            twisted_reactor)

        self._twisted_reactor = twisted_reactor

        def shutdown():
            gpsd.fini_service(gpsd_service)
            tz.fini_service(tz_service)
            fini_monitoring(stat_service)
            react.fini_service(react_service)
            about.fini_service(about_service)
            fini_service(tl_service)

        twisted_reactor.addSystemEventTrigger('before', 'shutdown', shutdown)

    def run(self):
        self._twisted_reactor.run()
