import functools

from edge import eventlog
from edge_bbox.types import mk_plate_bbox2, QuadI, RectF
from edge.lane_detector import LaneDetector
import math

from edge.image.image import ImEnv, ImEngine, ImSizeFitter
from edge.image.types import FileBlob, ImFooterVars
from edge.service.tz import ITzService
from edge.service.about import IAboutService
from edge.model.types import Dir

from .types import DetectorArgs, BusLaneEvidence, RuleArgs
from ..com.scene import SceneMask
from ..com.types import VideoArgs
from ..base.best import check_quad_dist_from_img_borders
from ..types import IDetector
from ...detectors.com.schedule import Schedule
from ...model.types import BaseEvent, Event, ViolationWaypoint, BestWaypoint, Vehicle, Waypoint, BestViolationWaypoint
from ...model.models import Track, TrTrackItem
from ...services.vs import IVsService, VsImageInfo, VsImageBufferError, VsError, VsSessionAbsentError


def _test_points_far_enought(quad1, quad2):
    """
    :type quad1: QuadI
    :type quad2: QuadI
    :rtype: bool
    """
    _, plate_h = quad1.wh
    max_distance = plate_h * 0.7  # % of plate height

    rect1_cx, rect1_cy = quad1.center
    rect2_cx, rect2_cy = quad2.center

    if math.fabs(rect1_cx - rect2_cx) < max_distance and math.fabs(rect1_cy - rect2_cy) < max_distance:
        # points are in same rect
        return False
    else:
        # points are far enought
        return True


def _filter_tr_items(tr_items, lr_scene_w, lr_scene_h):
    """
    :type tr_items: list[TrTrackItem]
    :type lr_scene_w: int
    :type lr_scene_h: int
    :rtype: list[TrTrackItem]
    """
    result = []
    # todo: lr_item can be null ?
    init_point = tr_items[0]  # type: TrTrackItem
    result.append(init_point)

    for index in range(1, len(tr_items), 1):
        fini_point = tr_items[index]  # type: TrTrackItem
        if fini_point.lr_item is None:
            continue

        if init_point.lr_item is None:
            continue

        if not check_quad_dist_from_img_borders(fini_point.lr_item.plate_quad, lr_scene_w, lr_scene_h):
            continue

        if _test_points_far_enought(init_point.lr_item.plate_quad, fini_point.lr_item.plate_quad):
            result.append(fini_point)
            init_point = fini_point

    return result


class _SceneMsRef(VsImageInfo):
    def __init__(self, tr_item):
        """
        :type tr_item: TrTrackItem
        """
        VsImageInfo.__init__(self, tr_item.lr_item.ts)
        self.tr_item = tr_item


class _RuleServices:
    def __init__(self, tz_service, vs_service, about_service, lane_service):
        """
        :type tz_service: ITzService
        :type vs_service: IVsService
        :type about_service: IAboutService
        :type lane_service: LaneDetector
        """
        self.tz = tz_service
        self.vs = vs_service
        self.about = about_service
        self.lane = lane_service


class Rule:
    def __init__(self, twisted_reactor, services, rule_args, detector_name, init_im_env, mid_im_env, fini_im_env,
                 best_im_env, video_conf):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorFromThreads
        :type services: _RuleServices
        :type rule_args: RuleArgs
        :type detector_name: str
        :type init_im_env: ImEnv
        :type mid_im_env: ImEnv
        :type fini_im_env: ImEnv
        :type best_im_env: ImEnv
        :type video_conf: VideoArgs | None
        """
        self._reactor = twisted_reactor
        self._detector_name = detector_name

        self._init_im_env = init_im_env
        self._mid_im_env = mid_im_env
        self._fini_im_env = fini_im_env
        self._best_im_env = best_im_env

        self._services = services
        self._scheduler = Schedule(rule_args.schedule, self._services.tz)
        self._args = rule_args

        self._video_conf = video_conf
        if video_conf:
            self._im_fitter = ImSizeFitter((video_conf.max_w, video_conf.max_h))

    def check(self, track, base, violation_tr_items):
        """
        :type track: Track
        :type base: BaseEvent
        :type violation_tr_items: list of TrTrackItem
        :rtype: Deferred | None
        """
        if base.best_wp.vehicle.vehicle_class not in self._args.limits.vehicle_types:
            return None

        # print 'rule %s' % self._args.name
        c_v_ts = violation_tr_items[0].ts
        l_v_ts = violation_tr_items[-1].ts

        if not (self._scheduler.check2(l_v_ts) or self._scheduler.check2(c_v_ts)):
            return None

        scene_ms_refs = map(_SceneMsRef, violation_tr_items)  # type: list[_SceneMsRef]
        init_scene_ms_refs = scene_ms_refs[0:5]
        # init_scene_ms_refs.sort(key=lambda ref: ref.ts)
        fini_scene_ms_refs = scene_ms_refs[:-6:-1]

        if track.lr_track.direction == Dir.APPROACHING:
            best_scene_ms_refs = fini_scene_ms_refs
        elif track.lr_track.direction == Dir.RECEDING:
            best_scene_ms_refs = init_scene_ms_refs
        else:
            best_scene_ms_refs = []

        center_ts = (c_v_ts + l_v_ts) * 0.5
        mid_scene_ms_refs = filter(lambda it: center_ts < it.ts < l_v_ts - 1e6, scene_ms_refs)

        # fini_scene_ms_refs.sort(key=lambda ref: ref.ts, reverse=True)

        try:
            init_winner_img = self._services.vs.export_image_rgba(self._init_im_env.scene_rgba, init_scene_ms_refs)
        except VsImageBufferError as ex:
            self._init_im_env.scene_rgba.reset(14, ex.min_w, ex.min_h, self._init_im_env.footer_h)
            init_winner_img = self._services.vs.export_image_rgba(self._init_im_env.scene_rgba, init_scene_ms_refs)

        try:
            mid_winner_img = self._services.vs.export_image_rgba(self._mid_im_env.scene_rgba, mid_scene_ms_refs)
        except VsImageBufferError as ex:
            self._mid_im_env.scene_rgba.reset(14, ex.min_w, ex.min_h, self._mid_im_env.footer_h)
            mid_winner_img = self._services.vs.export_image_rgba(self._mid_im_env.scene_rgba, mid_scene_ms_refs)

        try:
            fini_winner_img = self._services.vs.export_image_rgba(self._fini_im_env.scene_rgba, fini_scene_ms_refs)
        except VsImageBufferError as ex:
            self._fini_im_env.scene_rgba.reset(14, ex.min_w, ex.min_h, self._fini_im_env.footer_h)
            fini_winner_img = self._services.vs.export_image_rgba(self._fini_im_env.scene_rgba, fini_scene_ms_refs)

        try:
            best_winner_img = self._services.vs.export_image_rgba(self._best_im_env.scene_rgba, best_scene_ms_refs)
        except VsImageBufferError as ex:
            self._best_im_env.scene_rgba.reset(14, ex.min_w, ex.min_h, self._best_im_env.footer_h)
            best_winner_img = self._services.vs.export_image_rgba(self._best_im_env.scene_rgba, best_scene_ms_refs)

        if None in [init_winner_img, fini_winner_img, mid_winner_img, best_winner_img]:
            eventlog.log_d('fail: get init/fini/mid/best image from vs')
            return None

        plate_text = base.get_plate_text()
        if mid_winner_img.ts - init_winner_img.ts < 1e6:
            eventlog.log_i('fail: too short p-p interval (plate:{}, mid-to-inif:{:6.3}sec'.format(
                plate_text, (mid_winner_img.ts - init_winner_img.ts) / 1e6))
            return None

        if fini_winner_img.ts - mid_winner_img.ts < 1e6:
            eventlog.log_i('fail: too short p-p interval (plate:{}, fini-to-mid:{:6.3}sec'.format(
                plate_text, (fini_winner_img.ts - mid_winner_img.ts) / 1e6))
            return None

        return self._mk_evidence(base, track, init_winner_img.tr_item,
                                 mid_winner_img.tr_item, fini_winner_img.tr_item,
                                 best_winner_img.tr_item)

    def _mk_evidence(self, base_event, track, init_tr_item, mid_tr_item, fini_tr_item, best_tr_item):
        """
        :type init_tr_item: TrTrackItem
        :type mid_tr_item: TrTrackItem
        :type fini_tr_item: TrTrackItem
        :type best_tr_item: TrTrackItem
        :type track Track
        :type base_event: BaseEvent
        """
        plate_text = base_event.get_plate_text()

        # get best vehicle
        best_waypoint = base_event.best_wp  # type: BestWaypoint
        best_vehicle = best_waypoint.vehicle  # type: Vehicle

        def mk_waypoint(services, im_env, tr_item, w_type=Waypoint.NAME_VIOLATION):
            """
            :type services: _RuleServices
            :type im_env: ImEnv
            :type tr_item: TrTrackItem
            :type w_type: str

            :rtype: ViolationWaypoint

            :raise RuntimeError:
            """
            lr_scene_img = track.lr_track.best_image
            plate_rect_f = RectF(*tr_item.lr_item.plate_quad.get_rel_bbox(lr_scene_img.width, lr_scene_img.height))
            plate_bbox = mk_plate_bbox2(plate_rect_f, *im_env.scene_rgba.wh)

            waypoint_origin = services.about.get_origin(im_env.scene_rgba.ts)

            lane_set = self._services.lane.get_lane3(*tr_item.get_plate_quad().center)
            if len(lane_set) == 1:
                waypoint_origin.location.lane = lane_set.pop()

            waypoint_ts = im_env.scene_rgba.ts
            wp_direction = track.tr_track.get_item_direction(tr_item)

            footer_vars = ImFooterVars(
                services.tz, self._services.about, waypoint_ts, defs=im_env.footer_def_vars,
                location=waypoint_origin.location, geolocation=waypoint_origin.geolocation,
                plate_text=plate_text,
                spot_speed=tr_item.ra_item.speed if tr_item.ra_item else None,
                direction=wp_direction,
                radar_speed=best_vehicle.spot_speed, video_speed=best_vehicle.camera_speed,
                video_speed2=best_vehicle.video_speed2)

            waypoint = None
            if w_type == Waypoint.NAME_VIOLATION:
                im_blobs = im_env.engine.mk_blobs(
                    im_env.scene_rgba, plate_bbox, footer_vars, track.lr_track, tr_item.lr_item.plate_quad,
                    vehicle=True, scene=True, plate=True)

                if len(im_blobs) != 3:
                    eventlog.log_d('fail: get all blobs for evidence')
                    raise RuntimeError

                waypoint = ViolationWaypoint(
                    waypoint_ts, waypoint_origin.location,
                    position=None,
                    direction=wp_direction,
                    plate_quads=[tr_item.lr_item.plate_quad])

            elif w_type == Waypoint.NAME_VIOLATION_BEST:
                im_blobs = im_env.engine.mk_blobs(
                    im_env.scene_rgba, plate_bbox, footer_vars, track.lr_track, tr_item.lr_item.plate_quad,
                    vehicle=True, scene=True, plate=True, thumb=True)

                if len(im_blobs) != 4:
                    eventlog.log_d('fail: get all blobs for evidence')
                    raise RuntimeError

                waypoint = BestViolationWaypoint(
                    waypoint_ts, waypoint_origin.location,
                    position=None,
                    direction=wp_direction,
                    plate_quads=[tr_item.lr_item.plate_quad])
            else:
                eventlog.log_e('fail: wp not supported')
                raise RuntimeError

            waypoint.add_images(im_blobs)
            return waypoint

        try:
            init_wp = mk_waypoint(self._services, self._init_im_env, init_tr_item)
            mid_wp = mk_waypoint(self._services, self._mid_im_env, mid_tr_item)
            fini_wp = mk_waypoint(self._services, self._fini_im_env, fini_tr_item)
            best_wp = mk_waypoint(self._services, self._best_im_env, best_tr_item, Waypoint.NAME_VIOLATION_BEST)
        except RuntimeError:
            return None

        init_ts = self._init_im_env.scene_rgba.ts
        fini_ts = self._fini_im_env.scene_rgba.ts
        event_ts = fini_ts
        duration = int((fini_ts - init_ts) / 1000000)
        evidence = BusLaneEvidence(
            self._services.about.get_origin(event_ts), event_ts, duration,
            self._args, self._detector_name)
        evidence.add_waypoints([
            init_wp,
            mid_wp,
            fini_wp,
            best_wp
        ])

        return evidence

    def add_video(self, track, callback):
        """
        :type track: Track
        :type callback: (FileBlob | None) -> None
        :rtype: None
        """
        if self._video_conf:
            scaled_w, scaled_h = self._im_fitter.scaledown(
                self._init_im_env.scene_rgba.wh)

            # noinspection PyBroadException
            try:
                return self._services.vs.export_video(
                    track.lr_track.min_ts / 1000 - self._video_conf.pre_record,
                    track.lr_track.max_ts / 1000 + self._video_conf.post_record,
                    scaled_w, scaled_h, self._video_conf.quality,
                    callback)

            except VsSessionAbsentError:
                eventlog.log_w('fail: vs (session-absent)')

            except VsError:
                eventlog.log_e('fail: vs')

            except Exception:
                eventlog.log_x()

        callback(None)


def create(*args, **kwargs):
    """
    :rtype: IDetector
    """
    return _Detector(*args, **kwargs)


class _Detector(IDetector):
    def __init__(self, twisted_reactor,  tz_service, vs_service, about_service, lane_service, args):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorFromThreads
        :type vs_service: IVsService
        :type tz_service: ITzService
        :type lane_service: LaneDetector
        :type args: DetectorArgs
        """
        self._ruleNameToSheduler = {}
        self._lane_detector = lane_service
        self._scene = args.scene

        self._init_im_env = ImEnv(
            ImEngine(args.im_engine),
            args.im_engine.scene.footer.height,
            args.def_footer_mapping)

        self._mid_im_env = ImEnv(
            self._init_im_env.engine,
            self._init_im_env.footer_h,
            self._init_im_env.footer_def_vars)

        self._fini_im_env = ImEnv(
            self._init_im_env.engine,
            self._init_im_env.footer_h,
            self._init_im_env.footer_def_vars)

        self._best_im_env = ImEnv(
            self._init_im_env.engine,
            self._init_im_env.footer_h,
            self._init_im_env.footer_def_vars)

        self._rules = []
        rule_services = _RuleServices(tz_service, vs_service, about_service, lane_service)
        for rule in args.rules:
            if rule.enabled:
                self._rules.append(Rule(
                    twisted_reactor,
                    rule_services, rule,
                    args.name,
                    self._init_im_env,
                    self._mid_im_env,
                    self._fini_im_env,
                    self._best_im_env,
                    args.video))

        self._identity = args.identity
        self._handler = None
        IDetector.__init__(self)

    def set_handler(self, event_handler):
        """
        :type event_handler: (Event) -> None
        """
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def handle_evidence(self, base, evidence, video):
        """
        :type base: BaseEvent
        :type evidence: BusLaneEvidence
        :type video: FileBlob | None
        """
        if video:
            evidence.add_videos([video])
        self._handler(Event(self._identity, base, evidence))

    def analyze(self, track, base):
        """
        :type track: Track
        :type base: BaseEvent
        """
        plate_text = base.get_plate_text()
        if '*' in plate_text:
            eventlog.log_d('done: skip event due to asterisk in plate-text')
            return

        if (track.lr_track.max_ts - track.lr_track.min_ts) / 1e6 < 2:
            eventlog.log_i("fail: to short track {} {:6.3} s".format(
                plate_text, (track.lr_track.max_ts - track.lr_track.min_ts) / 1e6))
            # too short trek
            return None

        sc_detector = SceneMask(self._scene)

        def get_parts():
            """
            :rtype: list of TrTrackItem
            """
            violation_tritem = []

            lr_scene_img = track.lr_track.best_image
            for tr_item in _filter_tr_items(track.tr_track, lr_scene_img.width, lr_scene_img.height):
            # for tr_item in track.tr_track:
                if tr_item.lr_item:
                    if not sc_detector.check_quad(tr_item.lr_item.plate_quad):
                        continue

                    violation_tritem.append(tr_item)

            return violation_tritem

        violation_tr_items = get_parts()

        if len(violation_tr_items) < 3:
            return None

        c_v_ts = violation_tr_items[0].ts
        l_v_ts = violation_tr_items[-1].ts

        if (l_v_ts - c_v_ts) / 1e6 < 2:
            eventlog.log_i("fail: to short track {} {:6.3} s".format(plate_text, (l_v_ts - c_v_ts) / 1e6))
            return None

        for rule in self._rules:
            evidence = rule.check(track, base, violation_tr_items)
            if evidence:
                handler = functools.partial(self.handle_evidence, base, evidence)
                rule.add_video(track, handler)
                break
