from collections import defaultdict

from edge import eventlog
from edge.types import ImageRef
from edge.model.trend import LrTrend, LrTrendItem

from .services.react import SceneReactProducer, TrackReactProducer
from .services.event import SceneEventConsumer, TrackEventConsumer
from .services.stats import StatCounters

from _args import KernelArgs, LrUpscaleArgs
from _match_store import LrRaMatchStore
from _models import LrScene, LrTrack, RaScene, RaSceneItem, RaTrack, TrScene, TrTrack
from _ra_track_store import RaTrackStore


def _get_median_fl(ra_items):
    """
    :type ra_items: list[RaSceneItem]
    :rtype: (float, float)
    """
    ra_items.sort(key=lambda ra_item: ra_item.raw_f)
    index = len(ra_items) / 2
    median_f = ra_items[index].raw_f

    ra_items.sort(key=lambda ra_item: ra_item.raw_l)
    index = len(ra_items) / 2
    median_l = ra_items[index].raw_l

    return median_f, median_l


def _get_median_fl2(ra_items):
    """
    :type ra_items: list[RaSceneItem]
    :rtype: (float, float)
    """

    def _cmp(left, right):
        """
        :type left: RaSceneItem
        :type right: RaSceneItem
        """
        left_d = left.raw_f * left.raw_f + left.raw_l * left.raw_l
        right_d = right.raw_f * right.raw_f + right.raw_l * right.raw_l
        return right_d - left_d

    ra_items.sort(cmp=_cmp)
    index = len(ra_items) / 2
    median_f = ra_items[index].raw_f
    median_l = ra_items[index].raw_l

    return median_f, median_l


def _filter_tr_track(tr_track):
    """
    :type tr_track: TrTrack
    :rtype: TrTrack
    """
    if None in [tr_track.lr_track, tr_track.ra_track]:
        return tr_track

    range_tout = int(10e6)  # 10 sec
    lost_tout = int(120e6)  # 2 min

    trend = LrTrend()
    trend.prolong(tr_track.lr_track.plate_trace, range_tout, lost_tout)

    ra_items = []
    ts_to_ra_items_map = defaultdict(list)
    for ra_item in tr_track.ra_track.trace:
        ra_item_processed = False
        for trend_item in trend:  # type: LrTrendItem
            if ra_item.ts in trend_item.range:
                ts_to_ra_items_map[trend_item.range.min_ts].append(ra_item)
                ra_item_processed = True
                break

        if not ra_item_processed:
            ra_items.append(ra_item)

    for items in ts_to_ra_items_map.itervalues():
        median_fl = _get_median_fl(items)
        for ra_item in items:  # type: RaSceneItem
            ra_item.raw_f, ra_item.raw_l = median_fl
            ra_items.append(ra_item)

    ra_items.sort(key=lambda x: x.ts)
    ra_track = RaTrack.create(ra_items)
    return TrTrack(tr_track.lr_track, ra_track)


class BaseKernel:
    def __init__(self,
                 scene_react_producer,
                 track_react_producer,
                 scene_event_consumers,
                 track_event_consumer,
                 lr_ra_match_store,
                 ra_track_store,
                 min_ra_track_len,
                 lr_upscale,
                 systest=False,
                 plate_detected_mode=False):
        """
        :type scene_react_producer: SceneReactProducer
        :type track_react_producer: TrackReactProducer
        :type scene_event_consumers: list[SceneEventConsumer]
        :type track_event_consumer: TrackEventConsumer
        :type lr_ra_match_store: LrRaMatchStore
        :type ra_track_store: RaTrackStore
        :type min_ra_track_len: int
        :type lr_upscale: LrUpscaleArgs
        :type systest: bool
        :type plate_detected_mode: bool
        """
        self._min_ra_track_len = min_ra_track_len

        self._scene_react_producer = scene_react_producer
        self._scene_react_producer.set_handler(self._handle_scene_react)

        self._track_react_producer = track_react_producer
        self._track_react_producer.set_handler(self._handle_track_react)

        self._lr_ra_match_store = lr_ra_match_store
        self._ra_track_store = ra_track_store
        self._ra_track_store.set_tout_handler(self._handle_ra_track_tout)

        self._scene_event_consumers = scene_event_consumers
        self._track_event_consumer = track_event_consumer

        self._lr_upscale = lr_upscale
        self._systest = systest
        self._plate_detected_mode = plate_detected_mode

    def _handle_scene_react(self, im_ref, lr_scene, ra_scene):
        """
        :type im_ref: ImageRef | None
        :type lr_scene: LrScene | None
        :type ra_scene: RaScene | None
        """
        # Update _ra_track_store
        if ra_scene:
            self._ra_track_store.add_ra_scene(ra_scene)

        # Create the tr_scene from an avaliable lr_scene and ra_scene.
        if lr_scene and ra_scene:
            tr_scene = TrScene.create_lr_ra(
                self._lr_upscale, lr_scene.scene, ra_scene.scene,
                lr_scene.im_ref.width, lr_scene.im_ref.height, self._systest)
            self._lr_ra_match_store.add(lr_scene.source_id, tr_scene.tr_items)

        elif lr_scene:
            tr_scene = TrScene.create_lr(self._lr_upscale, lr_scene.scene)

        elif ra_scene:
            tr_scene = TrScene.create_ra(self._lr_upscale, ra_scene.scene)

        else:
            tr_scene = TrScene.create_empty(self._lr_upscale)

        if im_ref:
            for consumer in self._scene_event_consumers:
                consumer.consume_scene(im_ref, tr_scene)

    def _handle_track_react(self, lr_track):
        """
        :type lr_track: LrTrack
        """
        if not self._plate_detected_mode:
            if lr_track.status != LrTrack.STAT_LOST:
                eventlog.log_d('skip not valid-lost track')
                return

        ra_track = None
        ra_oids_desc = self._lr_ra_match_store.get_ra_oids(lr_track.source_id, lr_track.oids)
        for ra_oid in ra_oids_desc:
            ra_track = self._ra_track_store.try_get_ra_track(ra_oid)
            if ra_track:
                break

        if ra_track is None:
            tr_track = TrTrack(lr_track, None)

        elif len(ra_track) >= self._min_ra_track_len:
            self._ra_track_store.rm_ra_track(ra_track.oid)
            raw_tr_track = TrTrack(lr_track, ra_track)
            tr_track = _filter_tr_track(raw_tr_track)

        else:
            tr_track = TrTrack(lr_track, None)

        self._track_event_consumer.consume_track(tr_track, plate_text=lr_track.plate_text)

    def _handle_ra_track_tout(self, ra_track):
        """
        :type ra_track: RaTrack
        """
        if len(ra_track) >= self._min_ra_track_len:
            tr_track = TrTrack(None, ra_track)

            # noinspection PyBroadException
            try:
                self._track_event_consumer.consume_track(tr_track)
            except Exception:
                eventlog.log_x()


class Kernel(BaseKernel):
    def __init__(self, twisted_reactor, args):
        """
        :type args: KernelArgs
        """
        stats = StatCounters(
            twisted_reactor,
            scene_consumer_count=len(args.scene_event_args_list),
            stat_filepath=args.stat_counters_args.stat_filepath,
            stat_tout_sec=args.stat_counters_args.stat_tout_sec)

        track_react_producer = TrackReactProducer(
            twisted_reactor, args.track_react_args, stats.track_react)

        scene_react_producer = SceneReactProducer(
            twisted_reactor, args.scene_react_args, stats.scene_react)

        ident = 0
        scene_event_consumers = list()
        for scene_event_args in args.scene_event_args_list:
            consumer = SceneEventConsumer(ident, scene_event_args, stats.scene_event)
            scene_event_consumers.append(consumer)
            ident += 1

        track_event_consumer = TrackEventConsumer(
            twisted_reactor,
            args.track_event_args,
            stats.track_event,
            args.image_store_args)

        ra_track_store = RaTrackStore(twisted_reactor, args.ra_track_store_tout, args.systest)
        lr_ra_match_store = LrRaMatchStore(twisted_reactor, args.lr_ra_oid_match_store_tout)

        BaseKernel.__init__(
            self,
            scene_react_producer,
            track_react_producer,
            scene_event_consumers,
            track_event_consumer,
            lr_ra_match_store,
            ra_track_store,
            args.min_ra_track_len,
            args.lr_upscale_args,
            args.systest,
            args.plate_detected_mode)

        self._twisted_reactor = twisted_reactor
        self._stat_counters = stats

    def run(self):
        self._twisted_reactor.run()
